"""
ExtractAudioFeatures class and related helpers
"""

import numpy

from returnn.util.basic import CollectionReadCheckCovered


class ExtractAudioFeatures:
    """
    Currently uses librosa to extract MFCC/log-mel features.
    (Alternatives: python_speech_features, talkbox.features.mfcc, librosa)
    """

    def __init__(
        self,
        window_len=0.025,
        step_len=0.010,
        num_feature_filters=None,
        with_delta=False,
        norm_mean=None,
        norm_std_dev=None,
        features="mfcc",
        feature_options=None,
        random_permute=None,
        random_state=None,
        raw_ogg_opts=None,
        pre_process=None,
        post_process=None,
        sample_rate=None,
        num_channels=None,
        peak_normalization=True,
        preemphasis=None,
        join_frames=None,
    ):
        """
        :param float window_len: in seconds
        :param float step_len: in seconds
        :param int num_feature_filters:
        :param bool|int with_delta:
        :param numpy.ndarray|str|int|float|None norm_mean: if str, will interpret as filename, or "per_seq"
        :param numpy.ndarray|str|int|float|None norm_std_dev: if str, will interpret as filename, or "per_seq"
        :param str|function features: "mfcc", "log_mel_filterbank", "log_log_mel_filterbank", "raw", "raw_ogg"
        :param dict[str]|None feature_options: provide additional parameters for the feature function
        :param CollectionReadCheckCovered|dict[str]|bool|None random_permute:
        :param numpy.random.RandomState|None random_state:
        :param dict[str]|None raw_ogg_opts:
        :param function|None pre_process:
        :param function|None post_process:
        :param int|None sample_rate:
        :param int|None num_channels: number of channels in audio
        :param bool peak_normalization: set to False to disable the peak normalization for audio files
        :param float|None preemphasis: set a preemphasis filter coefficient
        :param int|None join_frames: concatenate multiple frames together to a superframe
        :return: float32 data of shape
        (audio_len // int(step_len * sample_rate), num_channels (optional), (with_delta + 1) * num_feature_filters)
        :rtype: numpy.ndarray
        """
        self.window_len = window_len
        self.step_len = step_len
        if num_feature_filters is None:
            if features == "raw":
                num_feature_filters = 1
            elif features == "raw_ogg":
                raise Exception("you should explicitly specify num_feature_filters (dimension) for raw_ogg")
            else:
                num_feature_filters = 40  # was the old default
        self.num_feature_filters = num_feature_filters
        self.preemphasis = preemphasis
        if isinstance(with_delta, bool):
            with_delta = 1 if with_delta else 0
        assert isinstance(with_delta, int) and with_delta >= 0
        self.with_delta = with_delta
        # join frames needs to be set before norm loading
        self.join_frames = join_frames
        if norm_mean is not None:
            if not isinstance(norm_mean, (int, float)):
                norm_mean = self._load_feature_vec(norm_mean)
        if norm_std_dev is not None:
            if not isinstance(norm_std_dev, (int, float)):
                norm_std_dev = self._load_feature_vec(norm_std_dev)
        self.norm_mean = norm_mean
        self.norm_std_dev = norm_std_dev
        if random_permute and not isinstance(random_permute, CollectionReadCheckCovered):
            random_permute = CollectionReadCheckCovered.from_bool_or_dict(random_permute)
        self.random_permute_opts = random_permute
        self.random_state = random_state
        self.features = features
        self.feature_options = feature_options
        self.pre_process = pre_process
        self.post_process = post_process
        self.sample_rate = sample_rate
        if num_channels is not None:
            assert self.features == "raw", "Currently, multiple channels are only supported for raw waveforms"
            self.num_dim = 3
        else:
            self.num_dim = 2
        self.num_channels = num_channels
        self.raw_ogg_opts = raw_ogg_opts
        self.peak_normalization = peak_normalization

    def _load_feature_vec(self, value):
        """
        :param str|None value:
        :return: shape (self.num_inputs,), float32
        :rtype: numpy.ndarray|str|None
        """
        if value is None:
            return None
        if isinstance(value, str):
            if value == "per_seq":
                return value
            value = numpy.loadtxt(value)
        assert isinstance(value, numpy.ndarray)
        assert value.shape == (self.get_feature_dimension(),)
        return value.astype("float32")

    def get_audio_features_from_raw_bytes(self, raw_bytes, seq_name=None):
        """
        :param io.BytesIO raw_bytes:
        :param str|None seq_name:
        :return: shape (time,feature_dim)
        :rtype: numpy.ndarray
        """
        if self.features == "raw_ogg":
            assert self.with_delta == 0 and self.norm_mean is None and self.norm_std_dev is None
            # We expect that raw_bytes comes from a Ogg file.
            try:
                from returnn.extern.ParseOggVorbis.returnn_import import ParseOggVorbisLib
            except ImportError:
                print("Maybe you did not clone the submodule extern/ParseOggVorbis?")
                raise
            return ParseOggVorbisLib.get_instance().get_features_from_raw_bytes(
                raw_bytes=raw_bytes.getvalue(), output_dim=self.num_feature_filters, **(self.raw_ogg_opts or {})
            )

        # Don't use librosa.load which internally uses audioread which would use Gstreamer as a backend,
        # which has multiple issues:
        # https://github.com/beetbox/audioread/issues/62
        # https://github.com/beetbox/audioread/issues/63
        # Instead, use PySoundFile, which is also faster. See here for discussions:
        # https://github.com/beetbox/audioread/issues/64
        # https://github.com/librosa/librosa/issues/681
        import soundfile  # noqa  # pip install pysoundfile

        # integer audio formats are automatically transformed in the range [-1,1]
        audio, sample_rate = soundfile.read(raw_bytes)
        return self.get_audio_features(audio=audio, sample_rate=sample_rate, seq_name=seq_name)

    def get_audio_features(self, audio, sample_rate, seq_name=None):
        """
        :param numpy.ndarray audio: raw audio samples, shape (audio_len,)
        :param int sample_rate: e.g. 22050
        :param str|None seq_name:
        :return: array (time,dim), dim == self.get_feature_dimension()
        :rtype: numpy.ndarray
        """
        if self.sample_rate is not None:
            assert sample_rate == self.sample_rate, "currently no conversion implemented..."

        if self.preemphasis:
            from scipy import signal  # noqa

            audio = signal.lfilter([1, -self.preemphasis], [1], audio)

        if self.peak_normalization:
            peak = numpy.max(numpy.abs(audio))
            if peak != 0.0:
                audio /= peak

        if self.random_permute_opts and self.random_permute_opts.truth_value:
            audio = _get_random_permuted_audio(
                audio=audio, sample_rate=sample_rate, opts=self.random_permute_opts, random_state=self.random_state
            )

        if self.pre_process:
            audio = self.pre_process(audio=audio, sample_rate=sample_rate, random_state=self.random_state)
            assert isinstance(audio, numpy.ndarray) and len(audio.shape) == 1

        if self.features == "raw":
            assert self.num_feature_filters == 1
            if audio.ndim == 1:
                audio = numpy.expand_dims(audio, axis=1)  # add dummy feature axis
            if self.num_channels is not None:
                if audio.ndim == 2:
                    audio = numpy.expand_dims(audio, axis=2)  # add dummy feature axis
                assert audio.shape[1] == self.num_channels
                assert audio.ndim == 3  # time, channel, feature
            feature_data = audio.astype("float32")

        else:
            kwargs = {
                "sample_rate": sample_rate,
                "window_len": self.window_len,
                "step_len": self.step_len,
                "num_feature_filters": self.num_feature_filters,
                "audio": audio,
            }

            if self.feature_options is not None:
                assert isinstance(self.feature_options, dict)
                kwargs.update(self.feature_options)

            if callable(self.features):
                feature_data = self.features(random_state=self.random_state, **kwargs)
            elif self.features == "mfcc":
                feature_data = _get_audio_features_mfcc(**kwargs)
            elif self.features == "log_mel_filterbank":
                feature_data = _get_audio_log_mel_filterbank(**kwargs)
            elif self.features == "log_log_mel_filterbank":
                feature_data = _get_audio_log_log_mel_filterbank(**kwargs)
            elif self.features == "db_mel_filterbank":
                feature_data = _get_audio_db_mel_filterbank(**kwargs)
            elif self.features == "linear_spectrogram":
                feature_data = _get_audio_linear_spectrogram(**kwargs)
            elif self.features == "f0":
                kwargs.pop("num_feature_filters")
                kwargs.pop("window_len")
                feature_data = _get_f0_values(**kwargs)
            else:
                raise Exception("non-supported feature type %r" % (self.features,))

        assert feature_data.ndim == self.num_dim, "got feature data shape %r" % (feature_data.shape,)
        assert feature_data.shape[-1] == self.num_feature_filters

        if self.with_delta:
            import librosa  # noqa

            deltas = [
                librosa.feature.delta(feature_data, order=i, axis=0).astype("float32")
                for i in range(1, self.with_delta + 1)
            ]
            feature_data = numpy.concatenate([feature_data] + deltas, axis=-1)
            assert feature_data.shape[1] == (self.with_delta + 1) * self.num_feature_filters

        if self.norm_mean is not None:
            if isinstance(self.norm_mean, str) and self.norm_mean == "per_seq":
                feature_data -= numpy.mean(feature_data, axis=0, keepdims=True)
            elif isinstance(self.norm_mean, (int, float)):
                feature_data -= self.norm_mean
            else:
                if self.num_dim == 2:
                    feature_data -= self.norm_mean[numpy.newaxis, :]
                elif self.num_dim == 3:
                    feature_data -= self.norm_mean[numpy.newaxis, numpy.newaxis, :]
                else:
                    assert False, "Unexpected number of dimensions: {}".format(self.num_dim)

        if self.norm_std_dev is not None:
            if isinstance(self.norm_std_dev, str) and self.norm_std_dev == "per_seq":
                feature_data /= numpy.maximum(numpy.std(feature_data, axis=0, keepdims=True), 1e-2)
            elif isinstance(self.norm_std_dev, (int, float)):
                feature_data /= self.norm_std_dev
            else:
                if self.num_dim == 2:
                    feature_data /= self.norm_std_dev[numpy.newaxis, :]
                elif self.num_dim == 3:
                    feature_data /= self.norm_std_dev[numpy.newaxis, numpy.newaxis, :]
                else:
                    assert False, "Unexpected number of dimensions: {}".format(self.num_dim)

        if self.join_frames is not None:
            pad_len = self.join_frames - (feature_data.shape[0] % self.join_frames)
            pad_len = pad_len % self.join_frames
            new_len = feature_data.shape[0] + pad_len
            if self.num_channels is None:
                new_shape = (new_len // self.join_frames, feature_data.shape[-1] * self.join_frames)
                pad_width = ((0, pad_len), (0, 0))
            else:
                new_shape = (new_len // self.join_frames, self.num_channels, feature_data.shape[-1] * self.join_frames)
                pad_width = ((0, pad_len), (0, 0), (0, 0))
            feature_data = numpy.pad(feature_data, pad_width=pad_width, mode="edge")
            feature_data = numpy.reshape(feature_data, newshape=new_shape, order="C")

        assert feature_data.shape[-1] == self.get_feature_dimension()
        if self.post_process:
            feature_data = self.post_process(feature_data, seq_name=seq_name)
            assert isinstance(feature_data, numpy.ndarray) and feature_data.ndim == self.num_dim
            assert feature_data.shape[-1] == self.get_feature_dimension()
        return feature_data

    def get_feature_dimension(self):
        """
        :rtype: int
        """
        return (self.with_delta + 1) * self.num_feature_filters * (self.join_frames or 1)


def _get_audio_linear_spectrogram(
    audio, sample_rate, window_len=0.025, step_len=0.010, num_feature_filters=512, center=True
):
    """
    Computes linear spectrogram features from an audio signal.
    Drops the DC component.

    :param numpy.ndarray audio: raw audio samples, shape (audio_len,)
    :param int sample_rate: e.g. 22050
    :param float window_len: in seconds
    :param float step_len: in seconds
    :param bool center: pads the signal with reflection so that the window center starts at 0.
    :return: (audio_len // int(step_len * sample_rate), num_feature_filters), float32
    :rtype: numpy.ndarray
    """
    import librosa  # noqa

    min_n_fft = int(window_len * sample_rate)
    assert num_feature_filters * 2 >= min_n_fft
    assert num_feature_filters % 2 == 0

    librosa_version = librosa.__version__.split(".")
    if int(librosa_version[0]) >= 1 or (int(librosa_version[0]) == 0 and int(librosa_version[1]) >= 9):
        stft_func = librosa.stft
    else:
        stft_func = librosa.core.stft
    spectrogram = numpy.abs(
        stft_func(
            audio,
            hop_length=int(step_len * sample_rate),
            win_length=int(window_len * sample_rate),
            n_fft=num_feature_filters * 2,
            center=center,
        )
    )

    # remove the DC part
    spectrogram = spectrogram[1:]

    assert spectrogram.shape[0] == num_feature_filters
    spectrogram = spectrogram.transpose().astype("float32")  # (time, dim)
    return spectrogram


def _get_audio_features_mfcc(
    audio,
    sample_rate,
    window_len=0.025,
    step_len=0.010,
    num_feature_filters=40,
    n_mels=128,
    fmin=0,
    fmax=None,
    center=True,
):
    """
    :param numpy.ndarray audio: raw audio samples, shape (audio_len,)
    :param int sample_rate: e.g. 22050
    :param float window_len: in seconds
    :param float step_len: in seconds
    :param int num_feature_filters: number of dct outputs to use
    :param int n_mels: number of mel filters
    :param int fmin: minimum frequency for mel filters
    :param int|None fmax: maximum frequency for mel filters (None -> use sample_rate/2)
    :param bool center: pads the signal with reflection so that the window center starts at 0.
    :return: (audio_len // int(step_len * sample_rate), num_feature_filters), float32
    :rtype: numpy.ndarray
    """
    import librosa  # noqa

    features = librosa.feature.mfcc(
        y=audio,
        sr=sample_rate,
        n_mfcc=num_feature_filters,
        n_mels=n_mels,
        fmin=fmin,
        fmax=fmax,
        hop_length=int(step_len * sample_rate),
        n_fft=int(window_len * sample_rate),
        center=center,
    )
    librosa_version = librosa.__version__.split(".")
    if int(librosa_version[0]) >= 1 or (int(librosa_version[0]) == 0 and int(librosa_version[1]) >= 7):
        rms_func = librosa.feature.rms
    else:
        rms_func = librosa.feature.rmse  # noqa
    energy = rms_func(
        y=audio, hop_length=int(step_len * sample_rate), frame_length=int(window_len * sample_rate), center=center
    )
    features[0] = energy  # replace first MFCC with energy, per convention
    assert features.shape[0] == num_feature_filters  # (dim, time)
    features = features.transpose().astype("float32")  # (time, dim)
    return features


def _get_audio_log_mel_filterbank(audio, sample_rate, window_len=0.025, step_len=0.010, num_feature_filters=80):
    """
    Computes log Mel-filterbank features from an audio signal.
    References:

      https://github.com/jameslyons/python_speech_features/blob/master/python_speech_features/base.py
      https://github.com/tensorflow/tensor2tensor/blob/master/tensor2tensor/data_generators/speech_recognition.py

    :param numpy.ndarray audio: raw audio samples, shape (audio_len,)
    :param int sample_rate: e.g. 22050
    :param float window_len: in seconds
    :param float step_len: in seconds
    :param int num_feature_filters:
    :return: (audio_len // int(step_len * sample_rate), num_feature_filters), float32
    :rtype: numpy.ndarray
    """
    import librosa  # noqa

    mel_filterbank = librosa.feature.melspectrogram(
        y=audio,
        sr=sample_rate,
        n_mels=num_feature_filters,
        hop_length=int(step_len * sample_rate),
        n_fft=int(window_len * sample_rate),
    )
    log_noise_floor = 1e-3  # prevent numeric overflow in log
    log_mel_filterbank = numpy.log(numpy.maximum(log_noise_floor, mel_filterbank))
    assert log_mel_filterbank.shape[0] == num_feature_filters
    log_mel_filterbank = log_mel_filterbank.transpose().astype("float32")  # (time, dim)
    return log_mel_filterbank


def _get_audio_db_mel_filterbank(
    audio,
    sample_rate,
    window_len=0.025,
    step_len=0.010,
    num_feature_filters=80,
    fmin=0,
    fmax=None,
    min_amp=1e-10,
    center=True,
):
    """
    Computes log Mel-filterbank features in dezibel values from an audio signal.
    Provides adjustable minimum frequency and minimual amplitude clipping

    :param numpy.ndarray audio: raw audio samples, shape (audio_len,)
    :param int sample_rate: e.g. 22050
    :param float window_len: in seconds
    :param float step_len: in seconds
    :param int num_feature_filters: number of mel-filterbanks
    :param int fmin: minimum frequency covered by mel filters
    :param int|None fmax: maximum frequency covered by mel filters
    :param int min_amp: silence clipping for small amplitudes
    :param bool center: pads the signal with reflection so that the window center starts at 0.
    :return: (audio_len // int(step_len * sample_rate), num_feature_filters), float32
    :rtype: numpy.ndarray
    """
    # noinspection PyPackageRequirements
    assert fmin >= 0
    assert min_amp > 0

    import librosa  # noqa

    mel_filterbank = librosa.feature.melspectrogram(
        y=audio,
        sr=sample_rate,
        n_mels=num_feature_filters,
        hop_length=int(step_len * sample_rate),
        n_fft=int(window_len * sample_rate),
        fmin=fmin,
        fmax=fmax,
        center=center,
    )

    log_mel_filterbank = 20 * numpy.log10(numpy.maximum(min_amp, mel_filterbank))
    assert log_mel_filterbank.shape[0] == num_feature_filters
    log_mel_filterbank = log_mel_filterbank.transpose().astype("float32")  # (time, dim)
    return log_mel_filterbank


def _get_audio_log_log_mel_filterbank(audio, sample_rate, window_len=0.025, step_len=0.010, num_feature_filters=80):
    """
    Computes log-log Mel-filterbank features from an audio signal.
    References:

      https://github.com/jameslyons/python_speech_features/blob/master/python_speech_features/base.py
      https://github.com/tensorflow/tensor2tensor/blob/master/tensor2tensor/data_generators/speech_recognition.py

    :param numpy.ndarray audio: raw audio samples, shape (audio_len,)
    :param int sample_rate: e.g. 22050
    :param float window_len: in seconds
    :param float step_len: in seconds
    :param int num_feature_filters:
    :return: (audio_len // int(step_len * sample_rate), num_feature_filters), float32
    :rtype: numpy.ndarray
    """
    import librosa  # noqa

    librosa_version = librosa.__version__.split(".")
    if int(librosa_version[0]) >= 1 or (int(librosa_version[0]) == 0 and int(librosa_version[1]) >= 9):
        db_func = librosa.amplitude_to_db
    else:
        db_func = librosa.core.amplitude_to_db
    mel_filterbank = librosa.feature.melspectrogram(
        y=audio,
        sr=sample_rate,
        n_mels=num_feature_filters,
        hop_length=int(step_len * sample_rate),
        n_fft=int(window_len * sample_rate),
    )
    log_noise_floor = 1e-3  # prevent numeric overflow in log
    log_mel_filterbank = numpy.log(numpy.maximum(log_noise_floor, mel_filterbank))
    log_log_mel_filterbank = db_func(log_mel_filterbank)
    assert log_log_mel_filterbank.shape[0] == num_feature_filters
    log_log_mel_filterbank = log_log_mel_filterbank.transpose().astype("float32")  # (time, dim)
    return log_log_mel_filterbank


def _get_random_permuted_audio(audio, sample_rate, opts, random_state):
    """
    :param numpy.ndarray audio: raw time signal
    :param int sample_rate:
    :param CollectionReadCheckCovered opts:
    :param numpy.random.RandomState random_state:
    :return: audio randomly permuted
    :rtype: numpy.ndarray
    """
    import librosa  # noqa
    import scipy.ndimage  # noqa
    import warnings

    audio = audio * random_state.uniform(opts.get("rnd_scale_lower", 0.8), opts.get("rnd_scale_upper", 1.0))
    if opts.get("rnd_zoom_switch", 1.0) > 0.0:
        opts.get("rnd_zoom_lower"), opts.get("rnd_zoom_upper"), opts.get("rnd_zoom_order")  # Mark as read.
    if random_state.uniform(0.0, 1.0) < opts.get("rnd_zoom_switch", 0.2):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # Alternative: scipy.interpolate.interp2d
            factor = random_state.uniform(opts.get("rnd_zoom_lower", 0.9), opts.get("rnd_zoom_upper", 1.1))
            audio = scipy.ndimage.zoom(audio, factor, order=opts.get("rnd_zoom_order", 3))  # noqa
    if opts.get("rnd_stretch_switch", 1.0) > 0.0:
        opts.get("rnd_stretch_lower"), opts.get("rnd_stretch_upper")  # Mark as read.
    if random_state.uniform(0.0, 1.0) < opts.get("rnd_stretch_switch", 0.2):
        rate = random_state.uniform(opts.get("rnd_stretch_lower", 0.9), opts.get("rnd_stretch_upper", 1.2))
        audio = librosa.effects.time_stretch(audio, rate=rate)
    if opts.get("rnd_pitch_switch", 1.0) > 0.0:
        opts.get("rnd_pitch_lower"), opts.get("rnd_pitch_upper", 1.0)  # Mark as read.
    if random_state.uniform(0.0, 1.0) < opts.get("rnd_pitch_switch", 0.2):
        n_steps = random_state.uniform(opts.get("rnd_pitch_lower", -1.0), opts.get("rnd_pitch_upper", 1.0))
        audio = librosa.effects.pitch_shift(audio, sr=sample_rate, n_steps=n_steps)
    opts.assert_all_read()
    return audio


def _get_f0_values(audio, sample_rate, step_len=0.010, fmin=0, fmax=None):
    """
    Calls pyworld dio and stonemask to retrieve f_0 values and replaces all NaN positions(silence) with 0
    Does not use librosa.pyin because it is a lot slower.

    :param numpy.ndarray audio: raw time signal
    :param int sample_rate: e.g. 22050
    :param float step_len: in seconds
    :param int fmin: minimum frequency covered by mel filters
    :param int|None fmax: maximum frequency covered by mel filters
    :rtype: numpy.ndarray
    :return: Pitch features for audio signal
    """
    import pyworld as pw  # noqa

    f_0, t = pw.dio(x=audio, fs=sample_rate, f0_floor=fmin, f0_ceil=fmax, frame_period=step_len * 1000)  # convert to ms
    f_0 = pw.stonemask(x=audio, f0=f_0, temporal_positions=t, fs=sample_rate)
    f_0 = numpy.nan_to_num(f_0, nan=0.0)
    f_0 = numpy.expand_dims(f_0, axis=1)  # (time, dim)
    return f_0



"""
Import ESPnet E-Branchformer model parameters
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Union
import returnn.frontend as rf
from returnn.frontend.encoder.e_branchformer import EBranchformerLayer, FeedForwardConvGated
from returnn.frontend.decoder.transformer import FeedForward

if TYPE_CHECKING:
    import torch
    from espnet2.asr.encoder.e_branchformer_encoder import EBranchformerEncoderLayer, ConvolutionalGatingMLP
    from espnet.nets.pytorch_backend.transformer.positionwise_feed_forward import (
        PositionwiseFeedForward,
    )
    from espnet.nets.pytorch_backend.transformer.attention import (
        MultiHeadedAttention,
        RelPositionMultiHeadedAttention,
    )


def import_params_espnet_e_branchformer_layer_to_rf(
    model_espnet: EBranchformerEncoderLayer, model_rf: EBranchformerLayer
):
    """
    Import params from ESPnet E-Branchformer layer to
    RF :class:`returnn.frontend.encoder.e_branchformer.EBranchformerLayer`.
    """
    from .torch_nn import (
        import_params_torch_conv1d_to_rf,
        import_params_torch_layer_norm_to_rf,
        import_params_torch_linear_to_rf,
    )
    from espnet2.asr.encoder.e_branchformer_encoder import EBranchformerEncoderLayer
    from espnet.nets.pytorch_backend.transformer.positionwise_feed_forward import (
        PositionwiseFeedForward,
    )

    assert isinstance(model_espnet, EBranchformerEncoderLayer)
    assert isinstance(model_rf, EBranchformerLayer)

    assert isinstance(model_espnet.feed_forward, PositionwiseFeedForward)
    assert isinstance(model_espnet.feed_forward_macaron, PositionwiseFeedForward)

    import_params_espnet_positionwise_feed_forward_to_rf(model_espnet.feed_forward_macaron, model_rf.ffn1)
    import_params_espnet_positionwise_feed_forward_to_rf(model_espnet.feed_forward, model_rf.ffn2)

    import_params_torch_layer_norm_to_rf(model_espnet.norm_ff_macaron, model_rf.ffn1_layer_norm)
    import_params_torch_layer_norm_to_rf(model_espnet.norm_ff, model_rf.ffn2_layer_norm)
    import_params_torch_layer_norm_to_rf(model_espnet.norm_mha, model_rf.self_att_layer_norm)
    import_params_torch_layer_norm_to_rf(model_espnet.norm_mlp, model_rf.cgmlp_layer_norm)
    import_params_torch_layer_norm_to_rf(model_espnet.norm_final, model_rf.final_layer_norm)

    # noinspection PyTypeChecker
    import_params_espnet_multi_headed_attention_to_rf(model_espnet.attn, model_rf.self_att)

    # noinspection PyTypeChecker
    import_params_espnet_convolutional_gating_mlp_to_rf(model_espnet.cgmlp, model_rf.cgmlp)

    import_params_torch_conv1d_to_rf(model_espnet.depthwise_conv_fusion, model_rf.merge.depthwise_conv_fusion)
    import_params_torch_linear_to_rf(model_espnet.merge_proj, model_rf.merge.merge_proj)

    num_params_espnet = 0
    for k, v in model_espnet.named_parameters():
        num_params_espnet += v.numel()
    num_params_rf = 0
    for k, v in model_rf.named_parameters():
        num_params_rf += v.num_elements()
    assert num_params_rf == num_params_espnet, f"num params RF {num_params_rf} != params ESPnet {num_params_espnet}"


def import_params_espnet_positionwise_feed_forward_to_rf(model_espnet: PositionwiseFeedForward, model_rf: FeedForward):
    """import"""
    from .torch_nn import import_params_torch_linear_to_rf

    assert model_rf.linear_ff.with_bias and model_rf.linear_out.with_bias
    import_params_torch_linear_to_rf(model_espnet.w_1, model_rf.linear_ff)
    import_params_torch_linear_to_rf(model_espnet.w_2, model_rf.linear_out)


def import_params_espnet_multi_headed_attention_to_rf(
    model_espnet: Union[MultiHeadedAttention, RelPositionMultiHeadedAttention],
    model_rf: Union[rf.SelfAttention, rf.RelPosSelfAttention],
):
    """import"""
    import torch
    from .torch_nn import import_params_torch_linear_to_rf
    from espnet.nets.pytorch_backend.transformer.attention import (
        MultiHeadedAttention,
        RelPositionMultiHeadedAttention,
    )

    assert isinstance(model_espnet, (MultiHeadedAttention, RelPositionMultiHeadedAttention))
    assert isinstance(model_rf, (rf.SelfAttention, rf.RelPosSelfAttention))
    assert model_espnet.h == model_rf.num_heads.dimension
    assert model_espnet.d_k == model_rf.key_dim_per_head.dimension
    dim = model_espnet.d_k * model_espnet.h
    nh = model_espnet.h
    hdim = dim // nh

    with torch.no_grad():
        # Torch Linear: (out,in), but RF has (in,out).
        q = model_espnet.linear_q.weight.T.reshape(dim, nh, hdim)  # (in,h,out/h)
        k = model_espnet.linear_k.weight.T.reshape(dim, nh, hdim)  # (in,h,out/h)
        v = model_espnet.linear_v.weight.T.reshape(dim, nh, hdim)  # (in,h,out/h)
        q_bias = model_espnet.linear_q.bias.reshape(nh, hdim)  # (h,out/h)
        k_bias = model_espnet.linear_k.bias.reshape(nh, hdim)  # (h,out/h)
        v_bias = model_espnet.linear_v.bias.reshape(nh, hdim)  # (h,out/h)
        qkv = torch.cat([q, k, v], dim=2)  # (in,h,out/h*3)
        qkv = qkv.reshape(dim, 3 * dim)  # (in,out*3)
        qkv_bias = torch.cat([q_bias, k_bias, v_bias], dim=1).reshape(dim * 3)  # (out*3,)
        model_rf.qkv.weight.raw_tensor.copy_(qkv)
        model_rf.qkv.bias.raw_tensor.copy_(qkv_bias)

        import_params_torch_linear_to_rf(model_espnet.linear_out, model_rf.proj)

        if isinstance(model_espnet, RelPositionMultiHeadedAttention):
            assert isinstance(model_rf, rf.RelPosSelfAttention)
            assert model_rf.linear_pos is not None
            assert model_rf.pos_bias_u is not None and model_rf.pos_bias_v is not None

            import_params_torch_linear_to_rf(model_espnet.linear_pos, model_rf.linear_pos)
            _reorder_rel_pos_emb_espnet_to_rf_(model_rf.linear_pos.weight.raw_tensor, dim=0)
            model_rf.pos_bias_u.raw_tensor.copy_(model_espnet.pos_bias_u)
            model_rf.pos_bias_v.raw_tensor.copy_(model_espnet.pos_bias_v)
        else:
            assert not isinstance(model_rf, rf.RelPosSelfAttention)

    num_params_espnet = 0
    for k, v in model_espnet.named_parameters():
        num_params_espnet += v.numel()
    num_params_rf = 0
    for k, v in model_rf.named_parameters():
        num_params_rf += v.num_elements()
    assert num_params_rf == num_params_espnet, f"num params RF {num_params_rf} != params ESPnet {num_params_espnet}"


def _reorder_rel_pos_emb_espnet_to_rf(x: torch.Tensor, *, dim=-1) -> torch.Tensor:
    if dim < 0:
        dim += x.ndim
    assert 0 <= dim < x.ndim
    if dim != x.ndim - 1:
        x = x.transpose(dim, -1)
    # x: [..., D]
    # x feat dims is sin/cos repeated after each other
    *o, d = x.shape
    x = x.reshape(*o, d // 2, 2)  # [..., D/2, 2]
    # PT goes over indices T-1,T-2,...,0,1,2,...,T-1.
    # RF goes the other way around.
    # We don't flip here, to show that a linear transformation of the features is also fine.
    # Flipping cos has no effect.
    # Flipping sin would be equivalent to negating the positional encoding.
    x[..., 0] = -x[..., 0]
    # RF has first the sin, then the cos.
    x = x.transpose(-1, -2).reshape(*o, d)  # [..., D]
    if dim != x.ndim - 1:  # transpose back
        x = x.transpose(dim, -1)
    return x


def _reorder_rel_pos_emb_espnet_to_rf_(x: torch.Tensor, *, dim=-1):
    import torch

    with torch.no_grad():
        x.copy_(_reorder_rel_pos_emb_espnet_to_rf(x, dim=dim))


def import_params_espnet_convolutional_gating_mlp_to_rf(
    model_espnet: ConvolutionalGatingMLP, model_rf: FeedForwardConvGated
):
    """import"""
    from .torch_nn import (
        import_params_torch_linear_to_rf,
        import_params_torch_layer_norm_to_rf,
        import_params_torch_conv1d_to_rf,
    )
    from espnet2.asr.encoder.e_branchformer_encoder import ConvolutionalGatingMLP

    assert isinstance(model_espnet, ConvolutionalGatingMLP)
    assert isinstance(model_rf, FeedForwardConvGated)

    import_params_torch_linear_to_rf(model_espnet.channel_proj1[0], model_rf.linear_ff)
    _reorder_espnet_cgmlp_linear_ff_to_rf_(model_rf.linear_ff.weight.raw_tensor)
    if model_rf.linear_ff.with_bias:
        _reorder_espnet_cgmlp_linear_ff_to_rf_(model_rf.linear_ff.bias.raw_tensor)
    import_params_torch_linear_to_rf(model_espnet.channel_proj2, model_rf.linear_out)
    import_params_torch_layer_norm_to_rf(model_espnet.csgu.norm, model_rf.norm)
    import_params_torch_conv1d_to_rf(model_espnet.csgu.conv, model_rf.conv)
    assert model_espnet.csgu.linear is None

    num_params_espnet = 0
    for k, v in model_espnet.named_parameters():
        num_params_espnet += v.numel()
    num_params_rf = 0
    for k, v in model_rf.named_parameters():
        num_params_rf += v.num_elements()
    assert num_params_rf == num_params_espnet, f"num params RF {num_params_rf} != params ESPnet {num_params_espnet}"


def _reorder_espnet_cgmlp_linear_ff_to_rf_(w: torch.Tensor):
    import torch

    dims = list(w.shape)
    with torch.no_grad():
        w.copy_(w.reshape(*dims[:-1], 2, dims[-1] // 2).flip(-2).reshape(*dims))



"""
Testing returnn.frontend.decoder.transformer.
"""

from __future__ import annotations

import _setup_test_env  # noqa
import sys
import unittest
import torch
from returnn.util import better_exchook
from returnn.util.debug import PyTracer, check_py_traces_rf_to_pt_equal
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


def _setup():
    try:
        import lovely_tensors

        lovely_tensors.monkey_patch()
    except ImportError:
        pass


_setup()


def test_llama():
    """
    Test that we can reproduce the Llama model.

    This here is the final complete test.
    There are several other sub-tests:

    - :func:`test_rotary_embedding`
    - :func:`test_rope_causal_self_att`

    Some references for the whole Llama model:
    https://github.com/meta-llama/llama/blob/main/llama/model.py
    https://github.com/meta-llama/llama3/blob/main/llama/model.py
    https://github.com/karpathy/llama2.c/blob/master/model.py
    https://github.com/huggingface/transformers/blob/main/src/transformers/models/llama/modeling_llama.py
    https://github.com/hkproj/pytorch-llama/blob/main/model.py
    https://github.com/likejazz/llama3.np/blob/main/llama3.py
    """
    from returnn.frontend.decoder.transformer import TransformerDecoder, TransformerDecoderLayer, FeedForwardGated
    from returnn.frontend.conversions.hf_llama import import_params_hf_llama_to_rf_transformer_decoder
    from transformers.models.llama.modeling_llama import LlamaDecoderLayer, LlamaForCausalLM, LlamaModel, LlamaConfig

    config = LlamaConfig(
        vocab_size=11,
        hidden_size=64,
        intermediate_size=64 * 4,
        num_hidden_layers=2,
        num_attention_heads=2,
        max_position_embeddings=128,
    )

    model_hf = LlamaForCausalLM(config)
    print("HF Model:")
    print(model_hf)
    print("Parameters:")
    num_params = 0
    for k, v in model_hf.named_parameters():
        print(f"{k}: {list(v.shape)} {v.dtype}")
        num_params += v.numel()
    print("Total number of parameters:", num_params)

    rf.select_backend_torch()

    model_dim = Dim(config.hidden_size, name="model")
    model_rf = TransformerDecoder(
        encoder_dim=None,
        vocab_dim=Dim(config.vocab_size, name="vocab"),
        model_dim=model_dim,
        num_layers=config.num_hidden_layers,
        pos_enc=None,
        norm=rf.RMSNorm,
        ff=FeedForwardGated,
        share_embedding=False,
        input_embedding_scale=1.0,
        decoder_layer_opts=dict(self_att=rf.RotaryPosCausalSelfAttention, self_att_opts=dict(with_bias=False)),
        num_heads=config.num_attention_heads,
        dropout=0,
        att_dropout=0,
    )
    print("RF Model:")
    print(model_rf)
    print("Parameters:")
    num_params = 0
    for k, v in model_rf.named_parameters():
        print(f"{k}: {list(v.dims)} {v.dtype}")
        num_params += v.num_elements()
    print("Total number of parameters:", num_params)

    import_params_hf_llama_to_rf_transformer_decoder(model_hf, model_rf)

    batch_dim = Dim(3, name="batch")
    seq_dim = Dim(rf.random_uniform([batch_dim], minval=7, maxval=13, dtype="int32"), name="seq")
    in_ = rf.random_uniform([batch_dim, seq_dim], sparse_dim=model_rf.vocab_dim)
    in_.name = "input_labels"

    with PyTracer([TransformerDecoder.__call__, TransformerDecoderLayer.__call__], Tensor) as trace_rf:
        out_rf, _ = model_rf(in_, spatial_dim=seq_dim, state=model_rf.default_initial_state(batch_dims=[batch_dim]))

    mask = rf.sequence_mask([batch_dim, seq_dim])
    with PyTracer([LlamaForCausalLM.forward, LlamaModel.forward, LlamaDecoderLayer.forward], torch.Tensor) as trace_hf:
        out_hf = model_hf(in_.raw_tensor, attention_mask=mask.raw_tensor)

    check_py_traces_rf_to_pt_equal(
        trace_rf.captured_locals,
        trace_hf.captured_locals,
        [
            (
                (TransformerDecoder.__call__, 0, "decoded", 0),
                (LlamaModel.forward, 0, "inputs_embeds", 0),
                (batch_dim, seq_dim, model_dim),
            ),
        ],
    )

    print("Check...")
    assert out_rf.raw_tensor.shape == out_hf.logits.shape
    torch.testing.assert_allclose(out_rf.raw_tensor, out_hf.logits)
    print("  all matched!")


def test_feed_forward_gated():
    from returnn.frontend.decoder.transformer import FeedForwardGated
    from returnn.frontend.conversions.hf_llama import import_params_hf_llama_mlp_to_rf_feed_forward_gated
    from transformers.models.llama.modeling_llama import LlamaMLP, LlamaConfig

    config = LlamaConfig(
        vocab_size=11,
        hidden_size=64,
        intermediate_size=64 * 4,
        num_hidden_layers=2,
        num_attention_heads=2,
        max_position_embeddings=128,
    )

    model_hf = LlamaMLP(config)

    rf.select_backend_torch()
    rf.set_random_seed(42)

    model_dim = Dim(config.hidden_size, name="model")
    model_rf = FeedForwardGated(out_dim=model_dim, ff_dim=Dim(config.intermediate_size, name="inter"), dropout=0.0)

    import_params_hf_llama_mlp_to_rf_feed_forward_gated(model_hf, model_rf)

    batch_dim = Dim(3, name="batch")
    seq_dim = Dim(rf.random_uniform([batch_dim], minval=7, maxval=13, dtype="int32"), name="seq")
    in_ = rf.random_uniform([batch_dim, seq_dim, model_dim])
    in_.name = "input"

    out_rf = model_rf(in_)
    out_rf = out_rf.copy_transpose((batch_dim, seq_dim, model_dim))

    out_hf = model_hf(in_.raw_tensor)

    print("Check...")
    assert out_rf.raw_tensor.shape == out_hf.shape
    torch.testing.assert_allclose(out_rf.raw_tensor, out_hf)
    print("  all matched!")


def test_transformer_rel_pos_att():
    """
    This tests that TransformerDecoder together with RelPosCausalSelfAttention
    and FeedForwardGated works in a reasonable standard setup.
    Works = does not cause exceptions.

    Additionally, we test an issue that dim tags seems to be leaking.
    """
    from returnn.tensor import TensorDict, batch_dim
    from returnn.frontend.decoder.transformer import TransformerDecoder, FeedForwardGated
    from returnn.datasets.util.vocabulary import Vocabulary
    from returnn.torch.data.extern_data import raw_dict_to_extern_data

    rf.select_backend_torch()

    vocab = Vocabulary.create_vocab_from_labels(
        [" ", "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k"], eos_label=0, bos_label=0
    )
    assert vocab.bos_label_id is not None and vocab.eos_label_id is not None
    vocab_dim = Dim(vocab.num_labels, name="vocab", vocab=vocab)

    model_def = rf.build_dict(
        TransformerDecoder,
        encoder_dim=None,
        num_layers=2,  # with higher number of layers, probelm occurs more, but also with 2
        model_dim=20,
        num_heads=2,
        pos_enc=None,
        norm=rf.build_dict(rf.RMSNorm),
        ff=rf.build_dict(FeedForwardGated),
        decoder_layer_opts=dict(self_att=rf.build_dict(rf.RelPosCausalSelfAttention, with_bias=False)),
        dropout=0.0,
        att_dropout=0.0,
    )
    model = rf.build_from_dict(model_def, vocab_dim=vocab_dim)
    assert isinstance(model, TransformerDecoder)

    leakages = []

    # Adapted from Dim reset_raw.
    def _num_referenced_dim_tags(self: Dim) -> int:
        visited = set()  # ids
        queue = [self]
        while queue:
            # noinspection PyShadowingNames
            dim: Dim = queue.pop()
            if id(dim) in visited:
                continue
            visited.add(id(dim))
            # noinspection PyProtectedMember
            dim_extra = dim._extra
            if dim_extra:
                # Any dims via dim math could also contain raw tensors,
                # so iterate through them.
                print("Dim:", dim)
                print(" cache_dim_math:", dim_extra.cache_dim_math)
                print(" same_as:", dim_extra.same_as)
                print(" copy_same_as:", dim_extra.copy_same_as)
                print(" same_for_batch_ctx:", dim_extra.same_for_batch_ctx)
                queue += dim_extra.cache_dim_math.values()
                if dim_extra.same_as:
                    queue.append(dim_extra.same_as)
                if dim_extra.copy_same_as:
                    queue.append(dim_extra.copy_same_as)
                queue += dim_extra.same_for_batch_ctx.values()
        print(f"{self} _num_referenced_dim_tags (reset_raw), visited {len(visited)}")
        return len(visited)

    time_dim = Dim(None, name="time")
    extern_data_template = TensorDict([Tensor("data", (batch_dim, time_dim), "int32", sparse_dim=vocab_dim)])

    prev_step_num_tags = 0
    for step in range(10):
        print("Step:", step)
        rf.init_train_step_run_ctx(train_flag=False, step=step)

        # Check that we don't have any dim tags leaking.
        # Do that right after init_train_step_run_ctx, because that might clean some previous caches.
        step_num_tags = _num_referenced_dim_tags(time_dim)
        if step > 1 and step_num_tags > prev_step_num_tags:
            leakages.append(step_num_tags - prev_step_num_tags)
        prev_step_num_tags = step_num_tags

        seq_lens = torch.randint(5, 11, (3,), dtype=torch.int32)
        extern_data = raw_dict_to_extern_data(
            {"data": torch.randint(0, vocab_dim.dimension, (3, seq_lens.max())), "data:seq_len": seq_lens},
            extern_data_template=extern_data_template,
            device="cpu",
        )

        targets = extern_data["data"]
        targets_spatial_dim = time_dim
        input_labels, (targets_w_eos_spatial_dim,) = rf.pad(
            targets, axes=[targets_spatial_dim], padding=[(1, 0)], value=vocab.bos_label_id
        )
        targets_w_eos, _ = rf.pad(
            targets,
            axes=[targets_spatial_dim],
            padding=[(0, 1)],
            value=vocab.eos_label_id,
            out_dims=[targets_w_eos_spatial_dim],
        )

        batch_dims = [batch_dim]

        # Gradients not relevant for this test.
        with torch.no_grad():
            logits, _ = model(
                input_labels,
                spatial_dim=targets_w_eos_spatial_dim,
                encoder=None,
                state=model.default_initial_state(batch_dims=batch_dims),
            )

            logits_packed, pack_dim = rf.pack_padded(
                logits, dims=batch_dims + [targets_w_eos_spatial_dim], enforce_sorted=False
            )
            targets_packed, _ = rf.pack_padded(
                targets_w_eos, dims=batch_dims + [targets_w_eos_spatial_dim], enforce_sorted=False, out_dim=pack_dim
            )

            log_prob = rf.log_softmax(logits_packed, axis=model.vocab_dim)
            # log_prob = rf.label_smoothed_log_prob_gradient(log_prob, 0.1, axis=model.target_dim)
            loss = rf.cross_entropy(
                target=targets_packed, estimated=log_prob, estimated_type="log-probs", axis=model.vocab_dim
            )
            loss.mark_as_loss("ce", use_normalized_loss=True)

            best = rf.reduce_argmax(logits_packed, axis=model.vocab_dim)
            frame_error = best != targets_packed
            frame_error.mark_as_loss(name="fer", as_error=True)

    assert not leakages, f"Leakages: {leakages}"


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute



#!/usr/bin/env python3

"""
This explicitly compiles some of the native ops, and will tell you the so-filenames.
Normally all native ops (e.g. NativeLstm2 etc) are compiled on-the-fly within RETURNN.
When you export the computation graph (e.g. via ``compile_tf_graph.py``),
you explicitly must load these native ops.
"""

from __future__ import annotations

import os
import sys
import typing

import _setup_returnn_env  # noqa
from returnn import __main__ as rnn
from returnn.log import log
import argparse
import returnn.util.basic as util


config = None  # type: typing.Optional["returnn.config.Config"]


def init(config_filename, log_verbosity):
    """
    :param str config_filename: filename to config-file
    :param int log_verbosity:
    """
    rnn.init_better_exchook()
    rnn.init_thread_join_hack()
    if config_filename:
        print("Using config file %r." % config_filename)
        assert os.path.exists(config_filename)
    rnn.init_config(config_filename=config_filename, command_line_options=[])
    global config
    config = rnn.config
    config.set("log", None)
    config.set("log_verbosity", log_verbosity)
    config.set("use_tensorflow", True)
    rnn.init_log()
    print("Returnn compile-native-op starting up.", file=log.v1)
    rnn.returnn_greeting()
    rnn.init_backend_engine()
    assert util.BackendEngine.is_tensorflow_selected(), "this is only for TensorFlow"
    rnn.init_faulthandler()
    if "network" in config.typed_dict:
        print("Loading network")
        from returnn.tf.network import TFNetwork

        network = TFNetwork(name="", config=config, rnd_seed=1, train_flag=False, eval_flag=True, search_flag=False)
        network.construct_from_dict(config.typed_dict["network"])


def main(argv):
    """
    Main entry.
    """
    from returnn.tf.util.basic import CudaEnv, OpCodeCompiler

    CudaEnv.verbose_find_cuda = True
    OpCodeCompiler.CollectedCompilers = []

    argparser = argparse.ArgumentParser(description="Compile some op")
    argparser.add_argument("--config", help="filename to config-file")
    argparser.add_argument("--native_op", help="op name. e.g. 'LstmGenericBase'")
    argparser.add_argument(
        "--blas_lib", default=None, help="specify which blas lib to use (path to .so or file name to search for)"
    )
    argparser.add_argument(
        "--search_for_numpy_blas",
        dest="search_for_numpy_blas",
        action="store_true",
        help="search for blas inside numpys .libs folder",
    )
    argparser.add_argument(
        "--no_search_for_numpy_blas",
        dest="search_for_numpy_blas",
        action="store_false",
        help="do not search for blas inside numpys .libs folder",
    )
    argparser.add_argument("--verbosity", default=4, type=int, help="5 for all seqs (default: 4)")
    argparser.add_argument("--output_file", help="if given, will write the list of libs to this file")
    args = argparser.parse_args(argv[1:])
    init(config_filename=args.config, log_verbosity=args.verbosity)

    import returnn.native_op as native_op
    from returnn.tf.native_op import make_op, OpMaker

    if args.native_op:
        print("Loading native op %r" % args.native_op)
        op_gen = getattr(native_op, args.native_op)
        assert issubclass(op_gen, native_op.NativeOpGenBase)
        make_op(
            op_gen,
            compiler_opts={"verbose": True},
            search_for_numpy_blas=args.search_for_numpy_blas,
            blas_lib=args.blas_lib,
        )

    libs = []
    if OpMaker.with_cuda and OpMaker.tf_blas_gemm_workaround:
        print("CUDA BLAS lib:", OpMaker.cuda_blas_gemm_so_filename())
        libs.append(OpMaker.cuda_blas_gemm_so_filename())
    elif OpMaker.with_cuda is False:
        print("No CUDA.")

    for compiler in OpCodeCompiler.CollectedCompilers:
        assert isinstance(compiler, OpCodeCompiler)
        print(compiler)
        # noinspection PyProtectedMember
        libs.append(compiler._so_filename)

    if libs:
        print("libs:")
        for fn in libs:
            print(fn)
    else:
        print("no libs compiled. use --native_op or --config")

    if args.output_file:
        with open(args.output_file, "w") as f:
            for fn in libs:
                f.write(fn + "\n")
        print("Wrote lib list to file:", args.output_file)


if __name__ == "__main__":
    main(sys.argv)



"""
Provides :class:`PostprocessingDataset`.
"""

from __future__ import annotations

from itertools import islice
from numpy.random import RandomState
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, TypeVar

from returnn.datasets.basic import DatasetSeq
from returnn.datasets.util.vocabulary import Vocabulary
from returnn.tensor import Tensor, TensorDict
from returnn.tensor.dim import Dim
from .basic import init_dataset
from .cached2 import CachedDataset2

__all__ = ["PostprocessingDataset", "LaplaceOrdering", "Sequential"]


class PostprocessingDataset(CachedDataset2):
    """
    A dataset that allows for generic post-processing of data from another dataset
    using a function on the segment level and on the level of multiple segments via
    an iterator.

    This allows integrating various data augmentation techniques like e.g. Mixup,
    SpecAugment or speed perturbation into the data loading pipeline.

    The integration into the data loading pipeline makes it easy to distribute the
    data processing work across multiple CPU cores using `MultiProcDataset` and in
    turn frees the GPU from data preprocessing tasks.

    Example usage::

        from returnn.tensor.dim import Dim, DimTypes

        time_dim = Dim(None, kind=DimTypes.Spatial)
        new_data_dim = Dim(128)

        train = {
            "class": "PostprocessingDataset",
            "dataset": {
                "class": "HDFDataset",
                "files": ["/path/to/data.hdf"],
            },
            # one of them, but not both:
            # (data: TensorDict, *, rng: numpy.random.RandomState, **kwargs) -> TensorDict
            "map_seq": map_seq,
            # (iter: Iterator[TensorDict], *, rng: numpy.random.RandomState, **kwargs) -> Iterator[TensorDict]
            "map_seq_stream": map_seqs,
            # only required when data shapes change wrt. the wrapped dataset:
            "map_outputs": {
                "data": {"dims": [time_dim, new_data_dim]},
            },
        }

    The dataset itself does not support its own seq ordering and relies on the wrapped
    dataset for seq ordering instead. Specifying a ``seq_ordering`` other than ``default``
    results in an error.

    However, we provide an iterator that implements the common `laplace:.NUM_SEQS_PER_BIN`-variant
    of seq ordering that any custom ``map_seq_stream``-style postprocessing iterator can be composed
    with to implement the ordering via :class:`LaplaceOrdering`.

    Like this::

        from returnn.datasets.postprocessing import LaplaceOrdering, Sequential

        def my_map_seq_stream(iterator):
            ...

        train = {
            "class": "PostprocessingDataset",
            # ...
            "map_seq_stream": Sequential(
                my_map_seq_stream,
                LaplaceOrdering(num_seqs_per_bin=1000),
            ),
        }
    """

    def __init__(
        self,
        dataset: Dict[str, Any],
        map_seq: Optional[Callable] = None,
        map_seq_stream: Optional[Callable] = None,
        map_outputs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """
        :param dataset: inner dataset to be post-processed
        :param map_seq: post processor function operating on the single-segment level.
            Signature: `(data: TensorDict, *, rng: numpy.random.RandomState, **kwargs) -> TensorDict`
            To avoid confusion on the order of how the processing functions are applied to the data, only one of
            ``map_seq`` and ``map_seq_stream`` can be specified at a time.
            To ensure forwards compatibility, the function must accept ``**kwargs`` as its last argument.
            This is enforced by passing randomly named parameters at runtime.
        :param map_seq_stream: post processor function operating on the multiple segment level via an iterator.
            Allows merging multiple segments into one, or generating multiple output segments from one input segment.
            Signature:
                ``(iter: Iterator[TensorDict], *, rng: numpy.random.RandomState, **kwargs) -> Iterator[TensorDict]``
            To avoid confusion on the order of how the processing functions are applied to the data, only one of
            ``map_seq`` and ``map_seq_stream`` can be specified at a time.
            To ensure forwards compatibility, the function must accept ``**kwargs`` as its last argument.
            This is enforced by passing randomly named parameters at runtime.
        :param map_outputs: Type and axis specification of the outputs of the mapping functions,
            like extern_data and model_outputs.
            To simplify the common case when no shapes change, this value can be left unspecified. The dataset then
            assumes the same data layout as returned by the wrapped dataset.
            Example: `map_outputs={"data": {"dim": 42}}`
        :param kwargs: see :class:`CachedDataset2`, :class:`Dataset`
        """
        super().__init__(**kwargs)

        if self.seq_ordering != "default":
            raise ValueError(f"{self}: specify seq_ordering in wrapped dataset, not in {self.__class__.__name__}")
        if map_seq is None and map_seq_stream is None:
            raise ValueError(f"{self}: need to either set map_seq or map_seq_stream")
        if map_seq and map_seq_stream:
            raise ValueError(f"{self}: cannot set both map_seq and map_seq_stream")

        self._dataset_def = dataset
        self._map_seq = map_seq
        self._map_seq_stream = map_seq_stream
        self._map_outputs = map_outputs
        self._rng = RandomState(self._get_random_seed_for_epoch(0))

        self._dataset = init_dataset(self._dataset_def, parent_dataset=self)
        if self._map_seq_stream is None:
            # if the stream mapper is set, the num_seqs may change and the estimation is less accurate
            self._estimated_num_seqs = self._dataset.estimated_num_seqs
        self._data_iter: Optional[Iterator[Tuple[int, TensorDict]]] = None

        self._in_tensor_dict_template = TensorDict(
            {name: self._make_tensor_template_from_input(name) for name in self._dataset.get_data_keys()}
        )
        if self._map_outputs is not None:
            self._out_tensor_dict_template = TensorDict()
            self._out_tensor_dict_template.update(self._map_outputs, auto_convert=True)
        else:
            self._out_tensor_dict_template = self._in_tensor_dict_template
        self.num_outputs = {
            k: (t.sparse_dim.size if t.sparse_dim else t.shape[-1] if len(t.shape) > 0 else 1, t.ndim)
            for k, t in self._out_tensor_dict_template.data.items()
        }
        self._default_input = "data" if "data" in self.num_outputs else next(iter(self.num_outputs.keys()))
        self.num_inputs = self.num_outputs[self._default_input][0]

        self.labels = {}
        for k, t in self._out_tensor_dict_template.data.items():
            if t.vocab:
                self.labels[k] = t.vocab.labels
            elif t.sparse_dim:  # sparse_dim but not vocab
                self.labels[k] = list(map(str, range(t.sparse_dim.dimension)))  # dummy labels

    def init_seq_order(
        self, epoch: Optional[int] = None, seq_list: Optional[List[str]] = None, seq_order: Optional[List[int]] = None
    ):
        """
        :param epoch:
        :param seq_list:
        :param seq_order:
        :return: whether the order changed (True is always safe to return)
        """
        super().init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)

        if epoch is None and seq_list is None and seq_order is None:
            self._num_seqs = 0
            return True

        self._rng = RandomState(self._get_random_seed_for_epoch(epoch=epoch))
        assert self._dataset is not None
        self._dataset.init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
        self._data_iter = enumerate(self._build_mapping_iter())
        return True

    def get_data_keys(self):
        """:return: available data keys"""
        return list(self._out_tensor_dict_template.data.keys())

    def get_data_dtype(self, key):
        """:return: dtype of data entry `key`"""
        return self._out_tensor_dict_template.data[key].dtype

    def _collect_single_seq(self, seq_idx: int) -> Optional[DatasetSeq]:
        while True:
            try:
                loaded_seq_idx, tensor_dict = next(self._data_iter)
            except StopIteration:
                return None
            assert loaded_seq_idx <= seq_idx, "_collect_single_seq must be done monotonically"
            if loaded_seq_idx != seq_idx:
                continue
            seq = DatasetSeq(features={k: t.raw_tensor for k, t in tensor_dict.data.items()}, seq_idx=seq_idx)
            return seq

    def _build_mapping_iter(self) -> Iterator[TensorDict]:
        """
        :return: an iterator applying both the segment level and across-segment transformations on the given dataset
        """

        def _validate_tensor_dict_iter(inner: Iterator[TensorDict]) -> Iterator[TensorDict]:
            for t_dict in inner:
                for data_key, out_t in self._out_tensor_dict_template.data.items():
                    in_t = t_dict.data[data_key]
                    assert (
                        in_t.ndim == out_t.batch_ndim
                        and in_t.dtype == out_t.dtype
                        and all(d.dimension in (d_, None) for (d, d_) in zip(in_t.dims, out_t.shape))
                    )
                yield t_dict

        data_iter = self._iterate_dataset()
        if self._map_seq_stream is not None:
            data_iter = self._map_seq_stream(
                data_iter, rng=self._rng, **{f"fwd_compatible_random_kwarg_{self._rng.randint(0, 1000)}": None}
            )
            assert isinstance(
                data_iter, Iterator
            ), f"map_seq_stream must produce an {Iterator.__name__}, but produced {type(data_iter).__name__}"
        return _validate_tensor_dict_iter(data_iter)

    def _iterate_dataset(self) -> Iterator[TensorDict]:
        """
        :return: generator providing data samples in the form of a TensorDict
        """
        data_keys = self._dataset.get_data_keys()

        seq_index = 0
        while self._dataset.is_less_than_num_seqs(seq_index):
            self._dataset.load_seqs(seq_index, seq_index + 1)
            tensor_dict = self._in_tensor_dict_template.copy_template()
            for data_key in data_keys:
                tensor_dict.data[data_key].raw_tensor = self._dataset.get_data(seq_index, data_key)
            if self._map_seq is not None:
                tensor_dict = self._map_seq(
                    tensor_dict, rng=self._rng, **{f"fwd_compatible_random_kwarg_{self._rng.randint(0, 1000)}": None}
                )
                assert isinstance(
                    tensor_dict, TensorDict
                ), f"map_seq must produce a {TensorDict.__name__}, but produced {type(tensor_dict).__name__}"
            yield tensor_dict
            seq_index += 1

    def _make_tensor_template_from_input(self, data_key: str) -> Tensor:
        dtype = self._dataset.get_data_dtype(data_key)
        if dtype == "string":
            dims = []
        else:
            feature_dims = [
                Dim(dimension=dim, name=f"{data_key}_dim{i + 1}")
                for i, dim in enumerate(self._dataset.get_data_shape(data_key))
            ]
            dims = [Dim(dimension=None, name=f"{data_key}_frame"), *feature_dims]
        sparse_dim = None
        if self._dataset.is_data_sparse(data_key):
            sparse_dim = Dim(dimension=self._dataset.get_data_dim(data_key), name=f"{data_key}_sparse")
            if data_key in self._dataset.labels:
                sparse_dim.vocab = Vocabulary.create_vocab_from_labels(self._dataset.labels[data_key])
        return Tensor(data_key, dims=dims, dtype=dtype, sparse_dim=sparse_dim)


class LaplaceOrdering(Callable[[Iterator[TensorDict]], Iterator[TensorDict]]):
    """
    Iterator compatible with :class:`PostprocessingDataset`'s ``map_seq_stream`` applying
    laplace sequence ordering based on the number of segments per bin.

    To be composed with any custom data postprocessing logic via :class:`Sequential`.
    """

    def __init__(self, num_seqs_per_bin: int, length_key: str = "data"):
        """
        :param num_seqs_per_bin: number of segments in a single laplace bin.
        :param length_key: data key to determine the segment length from for ordering.
        """
        self.length_key = length_key
        assert num_seqs_per_bin > 0
        self.num_seqs_per_bin = num_seqs_per_bin

    def __call__(self, iterator: Iterator[TensorDict], **kwargs) -> Iterator[TensorDict]:
        """:return: generator applying laplace sequence ordering on the data"""
        iterator = iter(iterator)
        is_down_phase = False

        while True:
            seq_buffer = list(islice(iterator, self.num_seqs_per_bin))
            seq_buffer.sort(key=self._get_seq_len, reverse=is_down_phase)
            yield from seq_buffer

            is_down_phase = not is_down_phase

            if len(seq_buffer) < self.num_seqs_per_bin:
                break

    def _get_seq_len(self, tdict: TensorDict) -> int:
        """
        :return: segment length of the segment in `tdict` as measured by `self.length_key` for comparison.
        """
        return tdict.data[self.length_key].raw_tensor.shape[0]


T = TypeVar("T", TensorDict, Iterator[TensorDict])


class Sequential:
    """
    Callable that composes multiple postprocessing functions into one by sequential application,
    i.e. Sequential(f, g)(x) = (g ∘ f)(x) = g(f(x)).

    Can either compose ``map_seq``-style single-segment processor functions or ``map_seq_stream``-style
    iterators operating on multiple segments. Just make sure not to mix both styles.
    """

    def __init__(self, *postprocessing_funcs: Callable):
        """
        :param postprocessing_funcs: Postprocessing functions to compose.
        """
        self.funcs = postprocessing_funcs

    def __call__(self, arg: T, **kwargs) -> T:
        """:return: result of sequential application of the postprocessing functions"""

        for func in self.funcs:
            arg = func(arg, **kwargs)
        return arg



# -*- coding: utf8 -*-

import _setup_test_env  # noqa
from nose.tools import assert_equal, assert_not_equal, assert_raises, assert_true, assert_is
from numpy.testing import assert_almost_equal
from returnn.util.basic import *
import sys
import os
import numpy as np
import numpy
import unittest
import textwrap
import signal

from returnn.util import better_exchook
from returnn.util.py_ext_mod_compiler import PyExtModCompiler

better_exchook.replace_traceback_format_tb()

my_dir = os.path.dirname(os.path.realpath(__file__))


def _sig_alarm_handler(signum, frame):
    raise Exception(f"Alarm (timeout) signal handler")


signal.signal(signal.SIGALRM, _sig_alarm_handler)


@contextlib.contextmanager
def timeout(seconds=10):
    """
    :param seconds: when the context is not closed within this time, an exception will be raised
    """
    signal.alarm(seconds)
    try:
        yield
    finally:
        signal.alarm(0)


def _get_tmp_dir() -> str:
    """
    :return: dirname
    """
    import tempfile
    import shutil
    import atexit

    name = tempfile.mkdtemp()
    assert name and os.path.isdir(name) and not os.listdir(name)
    atexit.register(lambda: shutil.rmtree(name))
    return name


def test_cmd_true():
    r = sys_cmd_out_lines("true")
    assert_equal(r, [])


def test_cmd_false():
    assert_raises(CalledProcessError, lambda: sys_cmd_out_lines("false"))


def test_cmd_stdout():
    r = sys_cmd_out_lines("echo 1; echo 2;")
    assert_equal(r, ["1", "2"])


def test_cmd_stderr():
    r = sys_cmd_out_lines("echo x >&2")
    assert_equal(r, [], "cmd() output should only cover stdout")


def test_hms():
    assert_equal(hms(5), "0:00:05")
    assert_equal(hms(65), "0:01:05")
    assert_equal(hms(65 + 60 * 60), "1:01:05")


def test_hms_fraction():
    assert_equal(hms_fraction(0, decimals=3), "0:00:00.000")
    assert_equal(hms_fraction(5, decimals=3), "0:00:05.000")
    assert_equal(hms_fraction(5.345, decimals=3), "0:00:05.345")
    assert_equal(hms_fraction(65.345, decimals=3), "0:01:05.345")


def test_uniq():
    assert (uniq(np.array([0, 1, 1, 1, 2, 2])) == np.array([0, 1, 2])).all()


def test_slice_pad_zeros():
    assert_equal(list(slice_pad_zeros(np.array([1, 2, 3, 4]), begin=1, end=3)), [2, 3])
    assert_equal(list(slice_pad_zeros(np.array([1, 2, 3, 4]), begin=-2, end=2)), [0, 0, 1, 2])
    assert_equal(list(slice_pad_zeros(np.array([1, 2, 3, 4]), begin=-2, end=6)), [0, 0, 1, 2, 3, 4, 0, 0])
    assert_equal(list(slice_pad_zeros(np.array([1, 2, 3, 4]), begin=2, end=6)), [3, 4, 0, 0])


def test_math_PiecewiseLinear():
    from returnn.util.math import PiecewiseLinear

    eps = 1e-5
    f = PiecewiseLinear({1: 2, 3: 4, 5: 1})
    assert_equal(f(0), 2)
    assert_equal(f(1 - eps), 2)
    assert_equal(f(1), 2)
    assert_almost_equal(f(1 + eps), 2, decimal=4)
    assert_equal(f(2), 3)
    assert_almost_equal(f(3 - eps), 4, decimal=4)
    assert_equal(f(3), 4)
    assert_almost_equal(f(3 + eps), 4, decimal=4)
    assert_equal(f(4), 2.5)
    assert_almost_equal(f(5 - eps), 1, decimal=4)
    assert_equal(f(5), 1)
    assert_equal(f(5 + eps), 1)
    assert_equal(f(6), 1)


def test_parse_orthography_into_symbols():
    assert_equal(list("hi"), parse_orthography_into_symbols("hi"))
    assert_equal(list(" hello "), parse_orthography_into_symbols(" hello "))
    assert_equal(list("  "), parse_orthography_into_symbols("  "))
    assert_equal(list("hello ") + ["[FOO]"] + list(" bar "), parse_orthography_into_symbols("hello [FOO] bar "))


def test_parse_orthography():
    assert_equal(list("hi ") + ["[HES]"] + list(" there") + ["[END]"], parse_orthography("hi [HES] there "))


def test_NumbersDict_minus_1():
    a = NumbersDict({"classes": 11, "data": 11})
    b = NumbersDict(10)
    r = a - b
    print(a, b, r)
    assert_equal(r, NumbersDict(numbers_dict={"classes": 1, "data": 1}, broadcast_value=-10))


def test_NumbersDict_eq_1():
    a = NumbersDict({"classes": 11, "data": 11})
    b = NumbersDict(11)
    r1 = a.elem_eq(b, result_with_default=False)
    r2 = a.elem_eq(b, result_with_default=True)
    r2a = a == b
    print(a, b, r1, r2, r2a)
    assert_is(all(r2.values()), r2a)
    assert_is(r1.value, None)
    assert_equal(r1.dict, {"classes": True, "data": True})
    assert_equal(r1, NumbersDict({"classes": True, "data": True}))
    assert_is(r2.value, None)
    assert_equal(r2.dict, {"classes": True, "data": True})
    assert_true(r2a)


def test_NumbersDict_eq_2():
    a = NumbersDict(10)
    assert_equal(a, 10)
    assert_not_equal(a, 5)


def test_NumbersDict_mul():
    a = NumbersDict(numbers_dict={"data": 3, "classes": 2}, broadcast_value=1)
    b = a * 2
    assert isinstance(b, NumbersDict)
    assert b.value == 2
    assert_equal(b.dict, {"data": 6, "classes": 4})


def test_NumbersDict_float_div():
    a = NumbersDict(numbers_dict={"data": 3.0, "classes": 2.0}, broadcast_value=1.0)
    b = a / 2.0
    assert isinstance(b, NumbersDict)
    assert_almost_equal(b.value, 0.5)
    assert_equal(list(sorted(b.dict.keys())), ["classes", "data"])
    assert_almost_equal(b.dict["data"], 1.5)
    assert_almost_equal(b.dict["classes"], 1.0)


def test_NumbersDict_int_floordiv():
    a = NumbersDict(numbers_dict={"data": 3, "classes": 2}, broadcast_value=1)
    b = a // 2
    assert isinstance(b, NumbersDict)
    assert_equal(b.value, 0)
    assert_equal(list(sorted(b.dict.keys())), ["classes", "data"])
    assert_equal(b.dict["data"], 1)
    assert_equal(b.dict["classes"], 1)


def test_NumbersDict_to_dict():
    a = NumbersDict(numbers_dict={"data": 3, "classes": 2}, broadcast_value=1)
    b = dict(a // 1)
    print(b)
    # We intentionally test whether the order is the same as before.
    assert list(b.keys()) == ["data", "classes"]


def test_collect_class_init_kwargs():
    class A(object):
        def __init__(self, a):
            pass

    class B(A):
        def __init__(self, b, **kwargs):
            super(B, self).__init__(**kwargs)
            pass

    class C(B):
        def __init__(self, b, c, **kwargs):
            super(C, self).__init__(**kwargs)
            pass

    kwargs = collect_class_init_kwargs(C)
    print(kwargs)
    assert_equal(sorted(kwargs), ["a", "b", "c"])


def test_terminal_size():
    terminal_size()


def test_try_get_caller_name():
    def sub():
        return try_get_caller_name()

    assert_equal(sub(), "test_try_get_caller_name")


def test_camel_case_to_snake_case():
    assert_equal(camel_case_to_snake_case("CamelCaseOp"), "camel_case_op")


def test_NativeCodeCompiler():
    native = NativeCodeCompiler(
        base_name="test_NativeCodeCompiler",
        code_version=1,
        code=textwrap.dedent(
            """\
            static int magic = 13;

            extern "C" void set_magic(int i) { magic = i; }
            extern "C" int get_magic() { return magic; }
            """
        ),
    )
    import ctypes

    lib = native.load_lib_ctypes()
    lib.set_magic.restype = None  # void
    lib.set_magic.argtypes = (ctypes.c_int,)
    lib.get_magic.restype = ctypes.c_int
    lib.get_magic.argtypes = ()

    assert_equal(lib.get_magic(), 13)
    lib.set_magic(42)
    assert_equal(lib.get_magic(), 42)


def test_PyExtModCompiler():
    mod_name = "TestPyExtModCompiler"
    native = PyExtModCompiler(
        base_name=mod_name,
        verbose=True,
        code_version=1,
        code=textwrap.dedent(
            f"""\
            #include <Python.h>

            static PyObject *
            func(PyObject *self, PyObject *args)
            {{
                int x;
                if (!PyArg_ParseTuple(args, "i", &x))
                    return NULL;
                return PyLong_FromLong(x + 42);
            }}

            static PyMethodDef _Methods[] = {{
                {{"func",  func, METH_VARARGS, "Func."}},
                {{NULL, NULL, 0, NULL}}        /* Sentinel */
            }};

            static struct PyModuleDef _Module = {{
                PyModuleDef_HEAD_INIT,
                "{mod_name}",   /* name of module */
                NULL,     /* module documentation, may be NULL */
                -1,       /* size of per-interpreter state of the module,
                             or -1 if the module keeps state in global variables. */
                _Methods
            }};

            PyMODINIT_FUNC
            PyInit_{mod_name}(void)
            {{
                return PyModule_Create(&_Module);
            }}
            """
        ),
    )
    lib_filename = native.get_lib_filename()
    print("lib_filename:", lib_filename)
    mod = native.load_py_module()

    assert_equal(mod.func(0), 42)
    assert_equal(mod.func(1), 43)
    assert_equal(mod.func(2), 44)
    assert_equal(mod.func(-2), 40)


def test_Stats():
    rnd = numpy.random.RandomState(42)
    m = rnd.uniform(-2.0, 10.0, (1000, 3))
    mean_ref = numpy.mean(m, axis=0)
    var_ref = numpy.var(m, axis=0)
    std_dev_ref = numpy.std(m, axis=0)
    print("ref mean/var/stddev:", mean_ref, var_ref, std_dev_ref)
    assert_almost_equal(numpy.sqrt(var_ref), std_dev_ref)
    stats = Stats()
    t = 0
    while t < len(m):
        s = int(rnd.uniform(10, 100))
        m_sub = m[t : t + s]
        print("sub seq from t=%i, len=%i" % (t, len(m_sub)))
        stats.collect(m_sub)
        t += s
    mean = stats.get_mean()
    std_dev = stats.get_std_dev()
    print("mean/stddev:", mean, std_dev)
    assert_almost_equal(mean, mean_ref)
    assert_almost_equal(std_dev, std_dev_ref)
    m -= mean[None, :]
    m /= std_dev[None, :]
    stats2 = Stats()
    t = 0
    while t < len(m):
        s = int(rnd.uniform(10, 100))
        m_sub = m[t : t + s]
        stats2.collect(m_sub)
        t += s
    mean0 = stats2.get_mean()
    stddev1 = stats2.get_std_dev()
    print("normalized mean/stddev:", mean0, stddev1)
    assert_almost_equal(mean0, 0.0)
    assert_almost_equal(stddev1, 1.0)


def test_deepcopy():
    deepcopy({"a": 1, "b": 2, "c": [3, {}, (), [42, True]]})


def test_deepcopy_mod():
    o = deepcopy({"a": 1, "b": 2, "c": sys})
    assert isinstance(o, dict)
    assert o["c"] is sys


def test_deepcopy_config():
    from returnn.config import Config

    config = Config()
    deepcopy(config)


def test_deepcopy_builtins():
    user_ns = {}
    custom_exec("", "<source.py>", user_ns, user_ns)
    print(user_ns)
    assert "__builtins__" in user_ns
    assert isinstance(user_ns["__builtins__"], dict)
    o = deepcopy(user_ns)
    assert o["__builtins__"] is user_ns["__builtins__"]  # no copy, directly reference this module dict


def test_get_func_kwargs():
    def dummy_func(net, var, update_ops):
        pass

    assert_equal(list(getargspec(dummy_func).args), ["net", "var", "update_ops"])


def test_next_type_attrib_in_mro_chain():
    class Base:
        def method(self):
            return id(self)

    class Foo(Base):
        def method(self):
            return 2

    class Bar(Foo):
        pass

    assert_equal(type_attrib_mro_chain(Foo, "method"), [Foo.method, Base.method])
    assert_equal(type_attrib_mro_chain(Bar, "method"), [Foo.method, Base.method])
    assert_equal(next_type_attrib_in_mro_chain(Bar, "method", Foo.method), Base.method)


def test_simple_obj_repr():
    class X:
        def __init__(self, a, b=13):
            self.a = a
            self.b = b

        __repr__ = simple_obj_repr

    x = X(a=42)
    x_repr = repr(x)

    assert_equal(x_repr, "X(a=42, b=13)")


def test_obj_diff_str():
    assert_equal(obj_diff_str({"a": 1, "b": 2}, {"a": 1, "b": 3}), "dict diff:\n['b'] self: 2 != other: 3")


def test_obj_diff_str_non_str_key():
    assert_equal(obj_diff_str({1: 1, 2: 2}, {1: 1, 2: 3}), "dict diff:\n[2] self: 2 != other: 3")


def test_obj_diff_list_allowed_mapping():
    a = {"a": 1, "b": {"A:a": 1, "A:b": 2}, "c": ["A:a"], "d": {"A:b"}}
    b = {"a": 1, "b": {"B:a": 1, "B:b": 2}, "c": ["B:a"], "d": {"B:b"}}
    c = {"a": 1, "b": {"B:a": 1, "B:b": 2}, "c": ["B:a"], "d": {"B:x"}}
    ab_diff = obj_diff_list(a, b)
    assert ab_diff
    for line in ab_diff:
        print(line)
    print("---")

    def _allowed_mapping(a_, b_):
        if isinstance(a_, str) and isinstance(b_, str):
            if a_.startswith("A:") and b_.startswith("B:"):
                return True
        return False

    ab_diff = obj_diff_list(a, b, allowed_mapping=_allowed_mapping)
    assert not ab_diff
    ac_diff = obj_diff_list(a, c, allowed_mapping=_allowed_mapping)
    assert ac_diff
    for line in ac_diff:
        print(line)
    assert_equal(
        ac_diff,
        [
            "dict diff:",
            "['b'] dict diff:",
            "['b']   key 'A:b' not in other",
            "['b']   key 'B:b' not in self",
        ],
    )


@unittest.skipIf(PY3, "only for Python 2")
def test_py2_utf8_str_to_unicode():
    assert_equal(py2_utf8_str_to_unicode("a"), "a")
    assert_is(type(py2_utf8_str_to_unicode("a")), str)
    assert_equal(py2_utf8_str_to_unicode("äöü"), "äöü")
    assert_is(type(py2_utf8_str_to_unicode("äöü")), unicode)


def test_CollectionReadCheckCovered():
    x = CollectionReadCheckCovered.from_bool_or_dict(True)
    assert x and x.truth_value
    x = CollectionReadCheckCovered.from_bool_or_dict(False)
    assert not x or not x.truth_value
    x = CollectionReadCheckCovered.from_bool_or_dict({})
    assert not x or not x.truth_value
    x = CollectionReadCheckCovered.from_bool_or_dict({"a": "b"})
    assert x and x.truth_value


def test_version():
    import returnn

    print("RETURNN version:", returnn.__version__, returnn.__long_version__)
    assert "1.0.0" not in returnn.__version__ and "unknown" not in returnn.__long_version__


def test_logging():
    # There is no real test. But you can interactively test this.
    import logging

    logging.getLogger("returnn").debug("Hello from returnn logger.")
    logging.getLogger("returnn.import_").debug("Hello from returnn.import_ logger.")


def test_import_():
    from returnn.import_ import import_

    mod = import_("github.com/rwth-i6/returnn-experiments", "common/test.py", "20210302-01094bef2761")
    print("Loaded mod %s, name %s, file %s" % (mod, mod.__name__, mod.__file__))
    assert_equal(mod.hello(), "hello world")


def test_import_root_repo_mod():
    from returnn.import_ import import_

    mod = import_("github.com/rwth-i6/returnn_common", "test.py", "20210602-1bc6822")
    print("Loaded mod %s, name %s, file %s" % (mod, mod.__name__, mod.__file__))
    assert_equal(mod.hello(), "hello world")


def test_import_root_repo_pkg():
    from returnn.import_ import import_

    mod = import_("github.com/rwth-i6/returnn_common", ".", "20210602-1bc6822")
    print("Loaded mod %s, name %s, file %s" % (mod, mod.__name__, mod.__file__))
    from returnn_import.github_com.rwth_i6.returnn_common.v20210602162042_1bc6822b2fd1 import test

    assert_equal(test.hello(), "hello world")


def test_import_root_repo_sub_mod():
    from returnn.import_ import import_

    mod = import_("github.com/rwth-i6/returnn_common", "test/hello.py", "20210603-3752d77")
    print("Loaded mod %s, name %s, file %s" % (mod, mod.__name__, mod.__file__))
    assert_equal(mod.hello(), "hello world")


def test_import_pkg_py_import():
    from returnn.import_ import import_

    mod = import_("github.com/rwth-i6/returnn-experiments", "common", "20210302-01094bef2761")
    print("Loaded mod %s, name %s, file %s" % (mod, mod.__name__, mod.__file__))
    # noinspection PyUnresolvedReferences
    from returnn_import.github_com.rwth_i6.returnn_experiments.v20210302133012_01094bef2761 import common

    assert common is mod


def test_import_wrong_date():
    from returnn.import_ import import_
    from returnn.import_.common import InvalidVersion

    try:
        import_("github.com/rwth-i6/returnn-experiments", "common/test.py", "20210301-01094bef2761")
    except InvalidVersion as exc:
        print("got expected exception:", exc)
    else:
        raise Exception("We expected an invalid version exception but got nothing.")


def test_import_wrong_pkg_py_import():
    from pprint import pprint
    from returnn.import_ import import_
    from returnn.import_.common import _registered_modules, MissingExplicitImport

    # Use some other commit here which is not used by the other tests, to not mess up.
    mod = import_("github.com/rwth-i6/returnn-experiments", "common", "20210302-10beb9d1c57e")
    print("Loaded mod %s, name %s, file %s" % (mod, mod.__name__, mod.__file__))
    print("Registered modules:")
    pprint(_registered_modules)

    # Mod name, without "common".
    mod_name = "returnn_import.github_com.rwth_i6.returnn_experiments.v20210302153450_10beb9d1c57e"
    assert mod_name in _registered_modules
    assert mod_name in sys.modules
    # Remove from both to force a reload.
    del _registered_modules[mod_name]
    del sys.modules[mod_name]

    try:
        import returnn_import.github_com.rwth_i6.returnn_experiments.v20210302153450_10beb9d1c57e
    except MissingExplicitImport as exc:  # but *not* a normal ImportError
        print("Got expected exception:", exc)
    else:
        raise Exception("We expected an import error exception but got nothing.")


def test_literal_py_to_pickle():
    import ast
    from returnn.util import literal_py_to_pickle

    def check(s):
        print("check:", s)
        a = ast.literal_eval(s)
        b = literal_py_to_pickle.literal_eval(s)
        assert_equal(a, b)

    checks = [
        "0",
        "1",
        "0.0",
        "1.23",
        '""',
        '"abc"',
        "''",
        "'abc'",
        '"abc\\n\\x00\\x01\\"\'abc"',
        "'hällö'",
        "'k\\u200eý'",
        "[]",
        "[1]",
        "[1,2,3]",
        "{}",
        "{'a': 'b', 1: 2}",
        "{1}",
        "{1,2,3}",
        "None",
        "{'a': None, 'b':1, 'c':None, 'd':'d'}",
    ]
    for s in checks:
        check(s)


def test_native_signal_handler():
    from returnn.util.debug import install_native_signal_handler

    install_native_signal_handler(reraise_exceptions=True)


class _PreInitUnpicklingRaiseException:
    def __getstate__(self):
        return 42

    def __setstate__(self, state):
        raise Exception("_PreInitUnpicklingRaiseException")


class _PreInitDummy:
    def __init__(self, *, payload):
        self.payload = payload

    def __call__(self):
        pass  # nothing


def _noop_func():
    pass


def test_NonDaemonicSpawnProcess_hang_1514():
    # https://github.com/rwth-i6/returnn/issues/1514
    from returnn.util.multi_proc_non_daemonic_spawn import NonDaemonicSpawnProcess

    with timeout():
        proc = NonDaemonicSpawnProcess(target=_noop_func)
        proc.start()
        proc.join()
        assert proc.exitcode == 0

        proc = NonDaemonicSpawnProcess(target=_noop_func)
        # Give it some payload large enough such that it will the pipe buffer to trigger the potential hang.
        proc.pre_init_func = _PreInitDummy(payload=["A" * 100_000, _PreInitUnpicklingRaiseException(), "B" * 100_000])
        # After fixing this, we expect that the unpickling of pre_init_func will raise an exception,
        # however, the parent then will not hang.
        proc.start()
        proc.join()
        assert proc.exitcode != 0


def test_expand_env_vars():
    os.environ["TMPA"] = "/testA"
    os.environ["TMPB"] = "testB"
    assert_equal(expand_env_vars("$TMPA/$TMPB/returnn/file_cache"), "/testA/testB/returnn/file_cache")


def test_bpe_PrefixTree():
    from returnn.util.bpe import PrefixTree, BpeOpts, BpePostMergeSymbol

    tree = PrefixTree(opts=BpeOpts(label_postfix_merge_symbol=BpePostMergeSymbol))
    tree.add("hello")
    tree.add("helo" + BpePostMergeSymbol)
    assert not tree.finished and not tree.bpe_finished
    assert set(tree.arcs.keys()) == {"h"}
    node = tree.arcs["h"]
    assert not node.finished and not node.bpe_finished
    assert set(node.arcs.keys()) == {"e"}
    node = node.arcs["e"]
    assert not node.finished and not node.bpe_finished
    assert set(node.arcs.keys()) == {"l"}
    node = node.arcs["l"]
    assert not node.finished and not node.bpe_finished
    assert set(node.arcs.keys()) == {"l", "o"}
    node_o = node.arcs["o"]
    assert not node_o.finished and node_o.bpe_finished
    # The following BpePostMergeSymbol arc and its node is somewhat arbitrary, probably does not matter...
    assert set(node_o.arcs.keys()) == {BpePostMergeSymbol}
    node_o_post = node_o.arcs[BpePostMergeSymbol]
    assert node_o_post.finished and not node_o_post.bpe_finished
    assert set(node_o_post.arcs.keys()) == set()


def test_bpe_PrefixTree_word_prefix():
    from returnn.util.bpe import PrefixTree, BpeOpts

    tree = PrefixTree(opts=BpeOpts(word_prefix_symbol="▁"))
    tree.add("▁hello")
    tree.add("▁hel")
    tree.add("lo")
    tree.add("▁hi")
    assert not tree.finished and not tree.bpe_finished
    assert set(tree.arcs.keys()) == {"▁", "l"}
    node = tree.arcs["▁"]
    assert not node.finished and not node.bpe_finished
    assert set(node.arcs.keys()) == {"h"}
    node = node.arcs["h"]
    assert not node.finished and not node.bpe_finished
    assert set(node.arcs.keys()) == {"e", "i"}
    node_ = node.arcs["i"]
    assert node_.finished and not node_.bpe_finished and not node_.arcs
    node = node.arcs["e"]
    assert not node.finished and not node.bpe_finished
    assert set(node.arcs.keys()) == {"l"}
    node = node.arcs["l"]
    assert node.finished and not node.bpe_finished
    assert set(node.arcs.keys()) == {"l"}
    node = node.arcs["l"]
    assert not node.finished and not node.bpe_finished
    assert set(node.arcs.keys()) == {"o"}
    node = node.arcs["o"]
    assert node.finished and not node.bpe_finished and not node.arcs


def test_bpe_DepthFirstSearch():
    import itertools
    from returnn.util.bpe import PrefixTree, BpeOpts, DepthFirstSearch

    tree = PrefixTree(opts=BpeOpts(label_postfix_merge_symbol="@@"))
    tree.add("llo")
    tree.add("helo@@")
    tree.add("he@@")

    dfs = DepthFirstSearch(tree, "hello")
    assert_equal(dfs.search(), ["he@@", "llo"])
    dfs = DepthFirstSearch(tree, "helo")
    assert_equal(dfs.search(), None)
    dfs = DepthFirstSearch(tree, "x")
    assert_equal(dfs.search(), None)
    dfs = DepthFirstSearch(tree, "llo")
    assert_equal(dfs.search(), ["llo"])

    tree.add("hello")
    dfs = DepthFirstSearch(tree, "hello")
    assert_equal(dfs.search(), ["hello"])
    dfs = DepthFirstSearch(tree, "hello", sampler=lambda: True)
    assert_equal(dfs.search(), ["he@@", "llo"])

    tree.add("hel@@")
    tree.add("lo")
    dfs = DepthFirstSearch(tree, "hello")
    assert_equal(dfs.search(), ["hello"])
    dfs = DepthFirstSearch(tree, "hello", sampler=lambda: True)
    assert_equal(dfs.search(), ["he@@", "llo"])
    dfs = DepthFirstSearch(tree, "hello", sampler=lambda _it=itertools.count(): next(_it) in {3})
    assert_equal(dfs.search(), ["hel@@", "lo"])


def test_bpe_DepthFirstSearch_word_prefix():
    import itertools
    from returnn.util.bpe import PrefixTree, BpeOpts, DepthFirstSearch

    tree = PrefixTree(opts=BpeOpts(word_prefix_symbol="▁"))
    tree.add("▁hello")
    tree.add("▁hel")
    tree.add("lo")

    search = DepthFirstSearch(tree, "hello")
    assert_equal(search.search(), ["▁hello"])
    search = DepthFirstSearch(tree, "hello", sampler=lambda: True)
    assert_equal(search.search(), ["▁hel", "lo"])

    tree.add("▁he")
    tree.add("llo")
    search = DepthFirstSearch(tree, "hello", sampler=lambda: True)
    assert_equal(search.search(), ["▁he", "llo"])
    search = DepthFirstSearch(tree, "hello", sampler=lambda _it=itertools.count(): next(_it) in {3})
    assert_equal(search.search(), ["▁hel", "lo"])


def test_bpe_CharSyncSearch():
    from returnn.util.bpe import PrefixTree, BpeOpts, CharSyncSearch

    tree = PrefixTree(opts=BpeOpts(label_postfix_merge_symbol="@@"))
    tree.add("llo")
    tree.add("helo@@")
    tree.add("he@@")

    search = CharSyncSearch(tree, "hello")
    assert_equal(search.search(), [["he@@", "llo"]])
    search = CharSyncSearch(tree, "helo")
    assert_equal(search.search(), [])
    search = CharSyncSearch(tree, "x")
    assert_equal(search.search(), [])
    search = CharSyncSearch(tree, "llo")
    assert_equal(search.search(), [["llo"]])

    tree.add("hello")
    search = CharSyncSearch(tree, "hello")
    assert_equal(search.search(), [["he@@", "llo"], ["hello"]])


def test_bpe_CharSyncSearch_word_prefix():
    from returnn.util.bpe import PrefixTree, BpeOpts, CharSyncSearch

    tree = PrefixTree(opts=BpeOpts(word_prefix_symbol="▁"))
    tree.add("▁hello")
    tree.add("▁hel")
    tree.add("lo")
    tree.add("▁hi")

    search = CharSyncSearch(tree, "hello")
    assert_equal(search.search(), [["▁hel", "lo"], ["▁hello"]])
    search = CharSyncSearch(tree, "helo")
    assert_equal(search.search(), [])
    search = CharSyncSearch(tree, "x")
    assert_equal(search.search(), [])
    search = CharSyncSearch(tree, "lo")
    assert_equal(search.search(), [])
    search = CharSyncSearch(tree, "hi")
    assert_equal(search.search(), [["▁hi"]])


def test_file_cache():
    from returnn.util.file_cache import FileCache, CachedFile

    # First create some dummy data, to be used for the test below.
    src_dir = _get_tmp_dir() + "/returnn/file_cache_src"
    os.makedirs(src_dir, exist_ok=True)
    os.makedirs(src_dir + "/dirA/subdirB", exist_ok=True)
    os.makedirs(src_dir + "/dirC", exist_ok=True)
    with open(src_dir + "/dummy1.txt", "w") as f:
        f.write("Hello dummy1.txt\n")
    with open(src_dir + "/dirA/subdirB/dummy2.txt", "w") as f:
        f.write("Hello dummy2.txt\n")
    with open(src_dir + "/dirC/dummy3.txt", "w") as f:
        f.write("Hello dummy3.txt\n")

    cache_dir = _get_tmp_dir() + "/returnn/file_cache"
    cache = FileCache(cache_directory=cache_dir)
    cache._lock_timeout = 5  # for testing

    # Copy some dummy file.
    cached_fn1 = cache.get_file(src_dir + "/dummy1.txt")
    assert cached_fn1 == cache_dir + src_dir + "/dummy1.txt" and os.path.exists(cached_fn1)
    with open(cached_fn1) as f:
        assert f.read() == "Hello dummy1.txt\n"
    cached_fn2 = cache.get_file(src_dir + "/dirA/subdirB/dummy2.txt")
    assert cached_fn2 == cache_dir + src_dir + "/dirA/subdirB/dummy2.txt" and os.path.exists(cached_fn2)
    with open(cached_fn2) as f:
        assert f.read() == "Hello dummy2.txt\n"
    cached_fn3 = cache.get_file(src_dir + "/dirC/dummy3.txt")
    assert cached_fn3 == cache_dir + src_dir + "/dirC/dummy3.txt" and os.path.exists(cached_fn3)
    with open(cached_fn3) as f:
        assert f.read() == "Hello dummy3.txt\n"
    target_values = {
        cache_dir + src_dir + "/dummy1.txt": 1,
        cache_dir + src_dir + "/dirA/subdirB/dummy2.txt": 1,
        cache_dir + src_dir + "/dirC/dummy3.txt": 1,
    }
    for k, v in cache._touch_files_thread.files.items():
        assert FileCache._is_info_filename(k) or (k in target_values and target_values[k] == v)

    # Check config handle_cached_files_in_config.
    config, config_cached_files = cache.handle_cached_files_in_config(
        {"class": "Dataset", "files": [CachedFile(src_dir + "/dirA/subdirB/dummy2.txt")]}
    )
    assert config == {"class": "Dataset", "files": [cache_dir + src_dir + "/dirA/subdirB/dummy2.txt"]}
    assert config_cached_files == [cache_dir + src_dir + "/dirA/subdirB/dummy2.txt"]

    target_values = {
        cache_dir + src_dir + "/dummy1.txt": 1,
        cache_dir + src_dir + "/dirA/subdirB/dummy2.txt": 2,
        cache_dir + src_dir + "/dirC/dummy3.txt": 1,
    }
    for k, v in cache._touch_files_thread.files.items():
        assert FileCache._is_info_filename(k) or (k in target_values and target_values[k] == v)

    # Check that file is kept up-to-date until we release it.
    mtimes = set()
    for t in range(10):
        print(f"Sec {t}...")
        mtime = os.stat(cache_dir + src_dir + "/dirA/subdirB/dummy2.txt").st_mtime
        mtimes.add(mtime)
        assert 0 <= time.time() - mtime < 2
        time.sleep(1)
    assert len(mtimes) > 1, "mtime should change"

    # Now release the files.
    cache.release_files(
        [
            cache_dir + src_dir + "/dummy1.txt",
            cache_dir + src_dir + "/dirA/subdirB/dummy2.txt",
            cache_dir + src_dir + "/dirC/dummy3.txt",
        ]
    )
    cache.release_files([cache_dir + src_dir + "/dirA/subdirB/dummy2.txt"])  # was acquired twice
    assert len(dict(cache._touch_files_thread.files)) == 0
    time.sleep(5)
    mtime = os.stat(cache_dir + src_dir + "/dirA/subdirB/dummy2.txt").st_mtime
    assert 4 <= time.time() - mtime

    # Check cleanup mechanism.
    cache._cleanup_files_always_older_than_days = 4 / 60 / 60 / 24
    print("Cache older than secs:", cache._cleanup_files_always_older_than_days * 24 * 60 * 60)
    cleanup_timestamp_file = cache_dir + "/.recent_full_cleanup"
    try:
        last_full_cleanup = os.stat(cleanup_timestamp_file).st_mtime
    except FileNotFoundError:
        last_full_cleanup = float("-inf")
    print("Recent cleanup is ago:", time.time() - last_full_cleanup, "secs")
    # Note: After the first cleanup, there are still dirs maybe left over, because dirs are only cleaned
    # when their mtime is a older than some threshold.
    # However, when we just deleted some file inside, that would have updated the mtime of the dir.
    # This effectively means that we need several cleanup runs to clean up also the empty dirs.
    cache._lock_timeout = 0.05  # used for the threshold
    prev_count_dirs = None
    while True:
        time.sleep(0.1)
        os.unlink(cleanup_timestamp_file)  # reset
        time.sleep(0.1)
        cache.cleanup()
        count_files = 0
        count_dirs = 0
        for root, dirs, files in os.walk(cache_dir):
            # ignore cleanup timestamp file
            for fn in dirs + files:
                print(f"{root}/{fn}", "age:", time.time() - os.stat(f"{root}/{fn}").st_mtime, "sec")
            leftover_files = [f for f in files if "recent_full_cleanup" not in f]
            count_files += len(leftover_files)
            count_dirs += len(dirs)
        print("count files:", count_files, "count dirs:", count_dirs)
        assert count_files == 0
        if prev_count_dirs is None:
            assert count_dirs > 0
        else:
            assert count_dirs < prev_count_dirs
        prev_count_dirs = count_dirs
        if count_dirs == 0:
            break


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute
